#!/usr/bin/python3

from setuptools import setup

setup(
    name="dulwich",
    description="Python Git Library",
    long_description="""Dulwich does this and that.
And here is a list of bullet points:
* a
* and also b
* and c
""",
    version="0.1.1",
    author="Jelmer Vernooij",
    author_email="jelmer@jelmer.uk",
    url="https://www.dulwich.io/",
)
